#!/usr/bin/env python
# coding=utf8
# File: test.py
import gym
import sys
import l3c.mazeworld
from l3c.mazeworld import MazeTaskSampler

def test_2d_maze(max_iteration, task_type):
    print("Testing 2D Maze with task type: ", task_type)
    maze_env = gym.make("mazeworld-2D-v0", max_steps=200, enable_render=False, view_grid=1, task_type=task_type)
    n = 9
    task = MazeTaskSampler(n=n, step_reward=-0.01, goal_reward=1.0)
    maze_env.set_task(task)
    iteration = 0
    while iteration < max_iteration:
        iteration += 1
        maze_env.reset()
        done=False
        sum_reward = 0
        while not done:
            state, reward, done, _ = maze_env.step(maze_env.action_space.sample())
            sum_reward += reward
        n += 2 # gradually increase the difficulty
        print("Get score %f, Increase the difficulty, n = %d"%(sum_reward, n))
        task = MazeTaskSampler(n=n, step_reward=-0.01, goal_reward=1.0)
        maze_env.set_task(task)

def test_3d_discrete_maze(max_iteration, task_type):
    print("Testing Discrete 3D Maze with task type: ", task_type)
    maze_env = gym.make("mazeworld-discrete-3D-v0", max_steps=200, enable_render=False, task_type=task_type)
    n = 9
    task = MazeTaskSampler(n=n, step_reward=-0.01, goal_reward=1.0)
    maze_env.set_task(task)
    iteration = 0
    while iteration < max_iteration:
        iteration += 1
        maze_env.reset()
        done=False
        sum_reward = 0
        while not done:
            state, reward, done, _ = maze_env.step(maze_env.action_space.sample())
            sum_reward += reward
        n += 2 # gradually increase the difficulty
        print("Get score %f, Increase the difficulty, n = %d"%(sum_reward, n))
        task = MazeTaskSampler(n=n, step_reward=-0.01, goal_reward=1.0)
        maze_env.set_task(task)

def test_3d_continuous_maze(max_iteration, task_type):
    print("Testing Continuous 3D Maze with task type: ", task_type)
    maze_env = gym.make("mazeworld-continuous-3D-v0", max_steps=1000, enable_render=False, task_type=task_type)
    n = 9
    task = MazeTaskSampler(n=n, step_reward=-0.001, goal_reward=1.0)
    maze_env.set_task(task)
    iteration = 0
    while iteration < max_iteration:
        iteration += 1
        maze_env.reset()
        done=False
        sum_reward = 0
        while not done:
            state, reward, done, _ = maze_env.step(maze_env.action_space.sample())
            sum_reward += reward
        n += 2 # gradually increase the difficulty
        print("Get score %f, Increase the difficulty, n = %d"%(sum_reward, n))
        task = MazeTaskSampler(n=n, step_reward=-0.001, goal_reward=1.0)
        maze_env.set_task(task)

if __name__=="__main__":
    test_2d_maze(10, "ESCAPE")
    test_2d_maze(10, "SURVIVAL")
    test_3d_discrete_maze(10, "ESCAPE")
    test_3d_discrete_maze(10, "SURVIVAL")
    test_3d_continuous_maze(10, "ESCAPE")
    test_3d_continuous_maze(10, "SURVIVAL")



import numpy
import pygame
import time
from numba import njit
from l3c.mazeworld.envs.dynamics import PI, PI_4

FAR_RGB = numpy.array([0, 0, 0], dtype="float32")
TRANSPARENT_RGB = numpy.array([0, 255, 0], dtype="float32")

@njit(cache=True)
def DDA_2D(pos, i, j, cell_number, cell_size, cos_ori, sin_ori, cell_walls, cell_transparent, max_vision):
    delta_dist_x = 1.0e+6 if abs(cos_ori) < 1.0e-6 else abs(cell_size / cos_ori)
    delta_dist_y = 1.0e+6 if abs(sin_ori) < 1.0e-6 else abs(cell_size / sin_ori)
    d_x = ((i + 1) * cell_size - pos[0]) if cos_ori > 0 else (i * cell_size - pos[0])
    d_y = ((j + 1) * cell_size - pos[1]) if sin_ori > 0 else (j * cell_size - pos[1])
    side_dist_x = 1.0e+6 if abs(cos_ori) < 1.0e-6 else d_x / cos_ori
    side_dist_y = 1.0e+6 if abs(sin_ori) < 1.0e-6 else d_y / sin_ori
    delta_i = 1 if(cos_ori > 0) else -1
    delta_j = 1 if(sin_ori > 0) else -1
    hit_i = i
    hit_j = j
    hit_dist = 0.0
    hit_side = 0
    hit_transparent_list = []
    if(cell_transparent[hit_i, hit_j] > 0.01):
        if(side_dist_x < side_dist_y):
            hit_transparent_list.append((side_dist_x, hit_i, hit_j, 0, cell_transparent[hit_i, hit_j]))
        else:
            hit_transparent_list.append((side_dist_y, hit_i, hit_j, 1, cell_transparent[hit_i, hit_j]))

    while hit_dist < max_vision:
        if(side_dist_x < side_dist_y):
            hit_i += delta_i
            side_dist_y -= side_dist_x
            hit_dist += side_dist_x
            if(hit_i < 0 or hit_i >= cell_number):
                if(hit_j < 0 or hit_j >= cell_number):
                    hit_dist = 1.0e+6
                    break
            else:
                if(cell_transparent[hit_i, hit_j] > 0.01):
                    hit_transparent_list.append((hit_dist, hit_i, hit_j, 0, cell_transparent[hit_i, hit_j]))
                if(cell_walls[hit_i, hit_j] > 0):
                    hit_side = 0
                    break
            side_dist_x = delta_dist_x
        else:
            hit_j += delta_j
            side_dist_x -= side_dist_y
            hit_dist += side_dist_y
            if(hit_i < 0 or hit_i >= cell_number):
                if(hit_j < 0 or hit_j >= cell_number):
                    hit_dist = 1.0e+6
                    break
            else:
                if(cell_transparent[hit_i, hit_j] > 0.01):
                    hit_transparent_list.append((hit_dist, hit_i, hit_j, 1, cell_transparent[hit_i, hit_j]))
                if(cell_walls[hit_i, hit_j] > 0):
                    hit_side = 1
                    break
            side_dist_y = delta_dist_y
    return hit_dist, hit_i, hit_j, hit_side, hit_transparent_list


@njit(cache=True)
def maze_view(pos, ori, vision_height, cell_walls, cell_transparent, cell_texts, cell_size, texture_array,
        ceil_text, ceil_height, text_size, max_vision, l_focal, vision_angle_h, resolution_h, resolution_v):
    vision_screen_half_size_h = numpy.tan(vision_angle_h / 2) * l_focal
    vision_screen_half_size_v = vision_screen_half_size_h * resolution_v / resolution_h
    pixel_size = 2.0 * vision_screen_half_size_h / resolution_h
    s_ori = numpy.sin(ori)
    c_ori = numpy.cos(ori)
    text_to_cell = text_size / cell_size
    max_cell_i = cell_walls.shape[0]
    max_cell_j = cell_walls.shape[1]
    pixel_factor = pixel_size / l_focal

    # prepare some maths
    rgb_array = numpy.zeros(shape=(resolution_h, resolution_v, 3), dtype="int32")
    transparent_array = numpy.zeros(shape=(resolution_h, resolution_v), dtype="float32")
    rgb_array[:, :] = FAR_RGB
    cos_hp_array = numpy.zeros(shape=(resolution_h,), dtype="float32")
    cos_abs_hp_array = numpy.zeros(shape=(resolution_h,), dtype="float32")
    sin_abs_hp_array = numpy.zeros(shape=(resolution_h,), dtype="float32")
    tan_hp = (- 0.5 - resolution_h / 2) * pixel_factor
    for d_h in range(resolution_h):
        tan_hp += pixel_factor
        cos_hp = numpy.sqrt(1.0 / (1.0 + tan_hp ** 2))
        sin_hp = tan_hp * cos_hp
        sin_abs_hp_array[d_h] = sin_hp * c_ori + cos_hp * s_ori
        cos_abs_hp_array[d_h] = cos_hp * c_ori - sin_hp * s_ori
        cos_hp_array[d_h] = cos_hp
    
    # paint floor
    for d_v in range(resolution_v - 1, resolution_v//2, -1):
        # horizontal distance on screen
        v_screen = (d_v + 0.5) * pixel_size - vision_screen_half_size_v
        distance = vision_height / v_screen * l_focal
        light_incident = v_screen / l_focal
        if(distance > max_vision):
            continue

        for d_h in range(resolution_h):
            eff_distance = distance / cos_hp_array[d_h]
            alpha = min(1.0, max(2.0 * eff_distance / max_vision - 1.0, 0.0)) * light_incident
            hit_x = eff_distance * cos_abs_hp_array[d_h] + pos[0]
            hit_y = eff_distance * sin_abs_hp_array[d_h] + pos[1]
            i = (hit_x / cell_size)
            j = (hit_y / cell_size)
            d_i = i - numpy.floor(i)
            d_j = j - numpy.floor(j)
            i = int(i)
            j = int(j)
            if(i < max_cell_i and i >= 0 and j < max_cell_j and j >= 0):
                text_id = cell_texts[i,j]
                d_i /= text_to_cell
                d_j /= text_to_cell
                d_i -= numpy.floor(d_i)
                d_j -= numpy.floor(d_j)
                d_i *= texture_array[text_id].shape[0]
                d_j *= texture_array[text_id].shape[1]
                rgb_array[d_h, d_v, :] = light_incident * (alpha * FAR_RGB + (1.0 - alpha) * texture_array[text_id][int(d_i), int(d_j)])
                if(cell_transparent[i, j] > 0.01):
                    transparent_factor = cell_transparent[i, j] * 0.50 + 0.10
                    rgb_array[d_h, d_v] = (1.0 - transparent_factor) * rgb_array[d_h, d_v] + transparent_factor * TRANSPARENT_RGB
                    transparent_array[d_h, d_v] = 1.0

    # paint ceil
    for d_v in range(resolution_v//2):
        v_screen = vision_screen_half_size_v - (d_v + 0.5) * pixel_size
        distance = (ceil_height - vision_height) / v_screen * l_focal
        light_incident = v_screen / l_focal
        if(distance > max_vision):
            continue

        for d_h in range(resolution_h):
            eff_distance = distance / cos_hp_array[d_h]
            alpha = min(1.0, max(2.0 * eff_distance / max_vision - 1.0, 0.0))
            hit_x = eff_distance * cos_abs_hp_array[d_h] + pos[0]
            hit_y = eff_distance * sin_abs_hp_array[d_h] + pos[1]
            t_i = int(hit_x / cell_size)
            t_j = int(hit_y / cell_size)
            i = (hit_x / text_size)
            j = (hit_y / text_size)
            d_i = i - numpy.floor(i)
            d_j = j - numpy.floor(j)
            d_i *= ceil_text.shape[0]
            d_j *= ceil_text.shape[1]
            rgb_array[d_h, d_v, :] = light_incident * (alpha * FAR_RGB + (1.0 - alpha) * ceil_text[int(d_i), int(d_j)])
            if(t_i >= 0 and t_i < max_cell_i and t_j >= 0 and t_j < max_cell_j and cell_transparent[t_i, t_j] > 0):
                transparent_factor = cell_transparent[t_i, t_j] * 0.50 + 0.10
                rgb_array[d_h, d_v] = (1.0 - transparent_factor) * rgb_array[d_h, d_v] + transparent_factor * TRANSPARENT_RGB
                transparent_array[d_h, d_v] = 1.0
    
    # paint wall
    for d_h in range(resolution_h):
        i = int(pos[0] / cell_size)
        j = int(pos[1] / cell_size)
        hit_dist, hit_i, hit_j, hit_side, hit_transparent = DDA_2D(
                pos, i, j, max_cell_i, cell_size, cos_abs_hp_array[d_h], sin_abs_hp_array[d_h], 
                cell_walls, cell_transparent, max_vision)
        if(hit_dist > max_vision):
            continue
        alpha = min(1.0, max(2.0 * hit_dist / max_vision - 1.0, 0.0))
        text_id = cell_texts[hit_i,hit_j]
        hit_pt_x = hit_dist * cos_abs_hp_array[d_h] + pos[0]
        hit_pt_y = hit_dist * sin_abs_hp_array[d_h] + pos[1]
        if(hit_side == 0): # hit vertical wall
           local_h = hit_pt_y / cell_size 
           local_h -= numpy.floor(local_h)
           light_incident = abs(cos_abs_hp_array[d_h])
        else:
           local_h = hit_pt_x / cell_size 
           local_h -= numpy.floor(local_h)
           light_incident = abs(sin_abs_hp_array[d_h])

        ratio = hit_dist * cos_hp_array[d_h] / l_focal
        top_v = (ceil_height - vision_height) / ratio
        bot_v = vision_height / ratio

        v_s = max(0, int((vision_screen_half_size_v - top_v) / pixel_size))
        v_e = min(resolution_v, int((vision_screen_half_size_v + bot_v) / pixel_size))

        for d_v in range(v_s, v_e):
            local_v = (vision_screen_half_size_v - (d_v + 0.5) * pixel_size) * ratio + vision_height
            d_i = local_h / text_size
            d_j = local_v / text_size
            d_i -= numpy.floor(d_i)
            d_j -= numpy.floor(d_j)
            d_i = int(texture_array[text_id].shape[0] * d_i)
            d_j = int(texture_array[text_id].shape[1] * d_j)
            rgb_array[d_h, d_v, :] = light_incident * (alpha * FAR_RGB + (1.0 - alpha) * texture_array[text_id][int(d_i), int(d_j)])

        # Add those transparent
        for hit_dist, hit_i, hit_j, hit_side, strength in hit_transparent:
            ratio = hit_dist * cos_hp_array[d_h] / l_focal
            transparent_factor = strength * 0.50 + 0.10
            top_v = (ceil_height - vision_height) / ratio
            bot_v = vision_height / ratio

            v_s = max(0, int((vision_screen_half_size_v - top_v) / pixel_size))
            v_e = min(resolution_v, int((vision_screen_half_size_v + bot_v) / pixel_size))
            for d_v in range(v_s, v_e):
                if(transparent_array[d_h, d_v] < 1):
                    rgb_array[d_h, d_v] = (1.0 - transparent_factor) * rgb_array[d_h, d_v] + transparent_factor * TRANSPARENT_RGB

    transparent_array =  numpy.expand_dims(transparent_array, axis=2)
        
    return rgb_array



"""
Core File of Maze Env
"""
import os
import numpy
import pygame
import random
from pygame import font
from numpy import random as npyrnd
from numpy.linalg import norm
from l3c.mazeworld.envs.maze_base import MazeBase

class MazeCore2D(MazeBase):
    def __init__(self, view_grid=2, task_type="SURVIVAL", max_steps=5000):
        super(MazeCore2D, self).__init__(
                view_grid=view_grid,
                task_type=task_type,
                max_steps=max_steps
                )

    def do_action(self, action):
        assert numpy.shape(action) == (2,)
        assert abs(action[0]) < 2 and abs(action[1]) < 2
        tmp_grid_i = self._agent_grid[0] + action[0]
        tmp_grid_j = self._agent_grid[1] + action[1]

        if(self._cell_walls[tmp_grid_i, tmp_grid_j] < 1):
            self._agent_grid[0] = tmp_grid_i
            self._agent_grid[1] = tmp_grid_j
        self._agent_loc = self.get_cell_center(self._agent_grid)

        reward, done = self.evaluation_rule()
        self.update_observation()
        return reward, done

    def visualize_obs(self):
        w,h = self._observation.shape
        c_w = w // 2
        c_h = h // 2
        obs_array = numpy.full((w,h,3), 255, dtype="int32")
        obs_array[numpy.where(self._observation < -0.50)] = numpy.asarray([0, 0, 0], dtype="int32")
        for idxes in numpy.argwhere(self._observation > 0.01):
            tidx = tuple(idxes)
            if(tidx==(c_w, c_h)):
                continue
            if(self.task_type == "SURVIVAL"):
                f = int(255 - 255 * self._observation[tidx])
            elif(self.task_type == "ESCAPE"):
                f = 0
            else:
                f = 255
            obs_array[tidx] = numpy.asarray([f, 255, f], dtype="int32")
        if(self.task_type == "SURVIVAL"):
            f = max(0, int(255 - 128 * self._life))
            obs_array[c_w, c_h] = numpy.asarray([255, f, f], dtype="int32")
        else:
            obs_array[c_w, c_h] = numpy.asarray([255, 0, 0], dtype="int32")

        obs_array = obs_array[:,::-1]
        return obs_array

    def render_observation(self):
        #Paint Observation
        empty_range = 40
        obs_surf = pygame.surfarray.make_surface(self.visualize_obs())
        obs_surf = pygame.transform.scale(obs_surf, (self._view_size - 2 * empty_range, self._view_size - 2 * empty_range))
        self._screen.blit(self._obs_logo,(5, 5))
        self._screen.blit(obs_surf, (empty_range, empty_range))

        pygame.draw.rect(self._screen, pygame.Color("blue"), 
                (empty_range, empty_range,
                self._view_size - 2 * empty_range, self._view_size - 2 * empty_range), width=1)
        pygame.draw.rect(self._screen, pygame.Color("red"), 
                (self._agent_grid[0] * self._render_cell_size + self._view_size, self._view_size - (self._agent_grid[1] + 1) * self._render_cell_size,
                self._render_cell_size, self._render_cell_size), width=0)

    def movement_control(self, keys):
        #Keyboard control cases
        if keys[pygame.K_LEFT]:
            return (-1, 0)
        if keys[pygame.K_RIGHT]:
            return (1, 0)
        if keys[pygame.K_UP]:
            return (0, 1)
        if keys[pygame.K_DOWN]:
            return (0, -1)
        return (0, 0)

    def update_observation(self):
        #Add the ground first
        #Find Relative Cells
        self._observation = - numpy.ones(shape=(2 * self.view_grid + 1, 2 * self.view_grid + 1), dtype="float32")
        x_s = self._agent_grid[0] - self.view_grid
        x_e = self._agent_grid[0] + self.view_grid + 1
        y_s = self._agent_grid[1] - self.view_grid
        y_e = self._agent_grid[1] + self.view_grid + 1
        i_s = 0
        i_e = 2 * self.view_grid + 1
        j_s = 0
        j_e = 2 * self.view_grid + 1
        if(x_s < 0):
            i_s = -x_s
            x_s = 0
        if(x_e > self._n):
            i_e -= x_e - self._n
            x_e = self._n
        if(y_s < 0):
            j_s = -y_s
            y_s = 0
        if(y_e > self._n):
            j_e -= y_e - self._n
            y_e = self._n
        # Observation: -1 for walls, 1 for goals, > 0 for foods
        self._observation[i_s:i_e, j_s:j_e] = - self._cell_walls[x_s:x_e, y_s:y_e]
        cur_goal = numpy.zeros((self._n, self._n))
        cur_goal[tuple(self._goal)] = 1.0
        if(self.task_type == "SURVIVAL"):
            self._observation[i_s:i_e, j_s:j_e] += self._cur_food_rewards[x_s:x_e, y_s:y_e]
            self._observation[self.view_grid, self.view_grid] = self._life
        elif(self.task_type == "ESCAPE"):
            self._observation[i_s:i_e, j_s:j_e] += cur_goal[x_s:x_e, y_s:y_e]



import gym
import l3c.mazeworld
from l3c.mazeworld import MazeTaskSampler

if __name__=='__main__':
    maze_env = gym.make("mazeworld-discrete-3D-v0", task_type="ESCAPE", max_steps=5000)
    n = 15
    task = MazeTaskSampler(n=n, cell_size=2.0, wall_height=3.2, allow_loops=True, crowd_ratio=0.35)
    maze_env.set_task(task)
    while True:
        maze_env.reset()
        done=False
        sum_reward = 0
        while not done:
            maze_env.render()
            state, reward, done, _ = maze_env.step(None)
            sum_reward += reward
        if(not maze_env.key_done):
            print("Episode is over! You got %.1f score."%sum_reward)
            maze_env.save_trajectory("trajectory_%dx%d.png"%(n, n))
            if(sum_reward > 0.0):
                n += 2 # gradually increase the difficulty
                print("Increase the difficulty, n = %d"%n)
            task = MazeTaskSampler(n=n, cell_size=2.0, wall_height=3.2, allow_loops=True, crowd_ratio=0.35)
            maze_env.set_task(task)
        else:
            break



"""
Gym Environment For Maze3D
"""
import numpy
import gym
import pygame

from gym import error, spaces, utils
from gym.utils import seeding
from l3c.mazeworld.envs.maze_2d import MazeCore2D
from l3c.mazeworld.envs.maze_continuous_3d import MazeCoreContinuous3D
from l3c.mazeworld.envs.maze_discrete_3d import MazeCoreDiscrete3D

DISCRETE_ACTIONS=[(-1, 0), (1, 0), (0, -1), (0, 1)]

class MazeWorldDiscrete3D(gym.Env):
    def __init__(self, 
            enable_render=True,
            render_scale=480,
            resolution=(320, 320),
            max_steps = 5000,
            task_type = "SURVIVAL"
            ):

        self.enable_render = enable_render
        self.render_viewsize = render_scale
        self.maze_core = MazeCoreDiscrete3D(
                resolution_horizon = resolution[0],
                resolution_vertical = resolution[1],
                max_steps = max_steps,
                task_type = task_type
                )

        # Turning Left/Right and go backward / forward
        self.action_space = spaces.Discrete(4)
        # observation is the x, y coordinate of the grid
        self.observation_space = spaces.Box(low=numpy.zeros(shape=(resolution[0], resolution[1], 3), dtype=numpy.float32), 
                high=numpy.full((resolution[0], resolution[1], 3), 256, dtype=numpy.float32),
                dtype=numpy.float32)

        self.need_reset = True
        self.need_set_task = True

    def set_task(self, task_config):
        self.maze_core.set_task(task_config)
        self.need_set_task = False

    def reset(self):
        if(self.need_set_task):
            raise Exception("Must call \"set_task\" before reset")
        state = self.maze_core.reset()
        if(self.enable_render):
            self.maze_core.render_init(self.render_viewsize)
            self.keyboard_press = pygame.key.get_pressed()
        self.need_reset = False
        self.key_done = False
        return state

    def step(self, action=None):
        if(self.need_reset):
            raise Exception("Must \"reset\" before doing any actions")
        reward = self.maze_core._step_reward

        if(action is None): # Only when there is no action input can we use keyboard control
            pygame.time.delay(100) # 10 FPS
            action = self.maze_core.movement_control(self.keyboard_press)
        else:
            action = DISCRETE_ACTIONS[action]
            
        reward, done = self.maze_core.do_action(action)
        if(done):
            self.need_reset=True
        info = {"steps": self.maze_core.steps}

        return self.maze_core.get_observation(), reward, done, info

    def render(self, mode="human"):
        if(mode != "human"):
            raise NotImplementedError("Only human mode is supported")
        self.key_done, self.keyboard_press = self.maze_core.render_update()

    def save_trajectory(self, file_name):
        self.maze_core.render_trajectory(file_name)

class MazeWorldContinuous3D(gym.Env):
    def __init__(self, 
            enable_render=True,
            render_scale=480,
            resolution=(320, 320),
            max_steps = 5000,
            task_type = "SURVIVAL"
            ):

        self.enable_render = enable_render
        self.render_viewsize = render_scale
        self.maze_core = MazeCoreContinuous3D(
                resolution_horizon = resolution[0],
                resolution_vertical = resolution[1],
                max_steps = max_steps,
                task_type = task_type
                )

        # Turning Left/Right and go backward / forward
        self.action_space = spaces.Box(low=numpy.array([-1.0, -1.0]), 
                high=numpy.array([1.0, 1.0]), dtype=numpy.float32)
        # observation is the x, y coordinate of the grid
        self.observation_space = spaces.Box(low=numpy.zeros(shape=(resolution[0], resolution[1], 3), dtype=numpy.float32), 
                high=numpy.full((resolution[0], resolution[1], 3), 256, dtype=numpy.float32),
                dtype=numpy.float32)

        self.need_reset = True
        self.need_set_task = True

    def set_task(self, task_config):
        self.maze_core.set_task(task_config)
        self.need_set_task = False

    def reset(self):
        if(self.need_set_task):
            raise Exception("Must call \"set_task\" before reset")
        state = self.maze_core.reset()
        if(self.enable_render):
            self.maze_core.render_init(self.render_viewsize)
            self.keyboard_press = pygame.key.get_pressed()
        self.need_reset = False
        self.key_done = False
        return state

    def step(self, action=None):
        if(self.need_reset):
            raise Exception("Must \"reset\" before doing any actions")
        reward = self.maze_core._step_reward
        if(action is None): # Only when there is no action input can we use keyboard control
            pygame.time.delay(20) # 50 FPS
            tr, ws = self.maze_core.movement_control(self.keyboard_press)
        else:
            tr = action[0]
            ws = action[1]
        reward, done = self.maze_core.do_action(tr, ws)

        if(done):
            self.need_reset=True
        info = {"steps": self.maze_core.steps}

        return self.maze_core.get_observation(), reward, done, info

    def render(self, mode="human"):
        if(mode != "human"):
            raise NotImplementedError("Only human mode is supported")
        self.key_done, self.keyboard_press = self.maze_core.render_update()

    def save_trajectory(self, file_name):
        self.maze_core.render_trajectory(file_name)

class MazeWorld2D(gym.Env):
    def __init__(self,
            enable_render=True,
            render_scale=480,
            max_steps = 5000,
            task_type = "SURVIVAL",
            view_grid = 2):
        self.enable_render = enable_render
        self.maze_core = MazeCore2D(view_grid=view_grid, max_steps=max_steps, task_type=task_type)
        self.render_viewsize = render_scale

        # Go EAST/WEST/SOUTH/NORTH
        self.action_space = spaces.Discrete(4)
        # observation is the x, y coordinate of the grid
        self.observation_space = spaces.Box(low=-1, high=1, shape=(3,3), dtype=numpy.int32)

        self.need_reset = True
        self.need_set_task = True

    def set_task(self, task_config):
        self.maze_core.set_task(task_config) 
        self.need_set_task = False

    def reset(self):
        if(self.need_set_task):
            raise Exception("Must call \"set_task\" before reset")
        state = self.maze_core.reset()
        if(self.enable_render):
            self.maze_core.render_init(self.render_viewsize)
            self.keyboard_press = pygame.key.get_pressed()
        self.need_reset = False
        self.key_done = False
        return state

    def step(self, action=None):
        if(self.need_reset):
            raise Exception("Must \"reset\" before doing any actions")
        if(action is None): # Only when there is no action input can we use keyboard control
            pygame.time.delay(100) # 10 FPS
            action = self.maze_core.movement_control(self.keyboard_press)
        else:
            action = DISCRETE_ACTIONS[action]
            
        reward, done = self.maze_core.do_action(action)

        if(done):
            self.need_reset=True
        info = {"steps": self.maze_core.steps}

        return self.maze_core.get_observation(), reward, done, info

    def render(self, mode="human"):
        if(mode != "human"):
            raise NotImplementedError("Only human mode is supported")
        self.key_done, self.keyboard_press = self.maze_core.render_update()

    def save_trajectory(self, file_name, additional=None):
        self.maze_core.render_trajectory(file_name, additional=additional)



"""
Core File of Maze Env
"""
import os
import numpy
import pygame
import random
import time
from pygame import font
from numpy import random as npyrnd
from numpy.linalg import norm

class MazeBase(object):
    def __init__(self, **kw_args):
        for k in kw_args:
            self.__dict__[k] = kw_args[k]
        pygame.init()

    def set_task(self, task_config):
        # initialize textures
        self._cell_walls = numpy.copy(task_config.cell_walls)
        self._cell_texts = task_config.cell_texts
        self._start = task_config.start
        self._n = numpy.shape(self._cell_walls)[0]
        self._goal = task_config.goal
        self._cell_size = task_config.cell_size
        self._wall_height = task_config.wall_height
        self._agent_height = task_config.agent_height
        self._step_reward = task_config.step_reward
        self._goal_reward = task_config.goal_reward
        self._food_rewards = task_config.food_rewards
        self._food_interval = task_config.food_interval
        self._max_life = task_config.max_life
        self._initial_life = task_config.initial_life

        assert self._agent_height < self._wall_height and self._agent_height > 0, "the agent height must be > 0 and < wall height"
        assert self._cell_walls.shape == self._cell_texts.shape, "the dimension of walls must be equal to textures"
        assert self._cell_walls.shape[0] == self._cell_walls.shape[1], "only support square shape"

    def reset(self):
        self._agent_grid = numpy.copy(self._start)
        self._agent_loc = self.get_cell_center(self._start)
        self._goal_loc = self.get_cell_center(self._goal)
        self._agent_trajectory = [numpy.copy(self._agent_grid)]

        # Maximum w and h in the space
        self._size = self._n * self._cell_size

        # Valid in 3D
        self._agent_ori = 0.0

        if(self.task_type == "SURVIVAL"):
            self._food_wait_refresh = numpy.zeros_like(self._food_rewards, dtype="int32")
            self._cur_food_rewards = numpy.copy(self._food_rewards)
            self._food_revival_count = numpy.copy(self._food_interval)
            self._life = self._initial_life
            self._cell_transparents = self._cur_food_rewards
        elif(self.task_type == "ESCAPE"):
            self._cell_transparents = numpy.zeros_like(self._cell_walls, dtype="int32")
            self._cell_transparents[self._goal] = 1.0
        self.update_observation()
        self.steps = 0
        return self.get_observation()

    def evaluation_rule(self):
        self.steps += 1
        self._agent_trajectory.append(numpy.copy(self._agent_grid))
        agent_grid_idx = tuple(self._agent_grid)

        if(self.task_type == "SURVIVAL"):
            if(self._cur_food_rewards[agent_grid_idx] > 1.0e-2):
                # Get the food
                reward = self._cur_food_rewards[agent_grid_idx]
                self._food_wait_refresh[agent_grid_idx] = 1
                self._cur_food_rewards[agent_grid_idx] = 0.0
            else:
                reward = 0.0
            self._life += reward + self._step_reward
            self._life = min(self._life, self._max_life)
            done = self._life < 0.0 or self.episode_is_over()

            # Refresh the food where necessary
            self._food_revival_count -= self._food_wait_refresh
            for idxes in numpy.argwhere(self._food_revival_count < 0):
                tidx = tuple(idxes)
                self._cur_food_rewards[tidx] = self._food_rewards[tidx]
                self._food_revival_count[tidx] = self._food_interval[tidx]
                self._food_wait_refresh[tidx] = 0

        elif(self.task_type == "ESCAPE"):
            goal = (tuple(self._goal) == agent_grid_idx)
            reward = self._step_reward + goal * self._goal_reward
            done = goal or self.episode_is_over() 

        return reward, done

    def do_action(self, action):
        raise NotImplementedError()

    def render_init(self, view_size):
        font.init()
        self._font = font.SysFont("Arial", 18)

        #Initialize the agent drawing
        self._render_cell_size = view_size / self._n
        self._view_size = view_size

        self._obs_logo = self._font.render("Observation", 0, pygame.Color("red"))

        self._screen = pygame.Surface((2 * view_size, view_size))
        self._screen = pygame.display.set_mode((2 * view_size, view_size))
        pygame.display.set_caption("RandomMazeRender - GodView")
        self._surf_god = pygame.Surface((view_size, view_size))
        self._surf_god.fill(pygame.Color("white"))
        it = numpy.nditer(self._cell_walls, flags=["multi_index"])
        for _ in it:
            x,y = it.multi_index
            if(self._cell_walls[x,y] > 0):
                pygame.draw.rect(self._surf_god, pygame.Color("black"), (x * self._render_cell_size, view_size - (y + 1) * self._render_cell_size,
                        self._render_cell_size, self._render_cell_size), width=0)
            if(self.task_type == "ESCAPE" and x == self._goal[0] and y == self._goal[1]):
                pygame.draw.rect(self._surf_god, pygame.Color("green"), (x * self._render_cell_size, view_size - (y + 1) * self._render_cell_size,
                        self._render_cell_size, self._render_cell_size), width=0)
        logo_god = self._font.render("GodView", 0, pygame.Color("red"))
        self._surf_god.blit(logo_god,(view_size - 90, 5))

    def draw_food(self, scr, offset):
        it = numpy.nditer(self._cell_walls, flags=["multi_index"])
        for _ in it:
            x,y = it.multi_index
            if(self._cur_food_rewards[x,y] > 1.0e-2):
                f = int(255 - 255 * self._cur_food_rewards[x,y])
                pygame.draw.rect(scr,  pygame.Color(f, 255, f),
                        (x * self._render_cell_size + offset[0], offset[1] + self._view_size - (y + 1) * self._render_cell_size,
                        self._render_cell_size, self._render_cell_size), width=0)
                txt_life = self._font.render("Life: %f"%self._life, 0, pygame.Color("red"))
                scr.blit(txt_life,(self._view_size + 90, 5))

    def render_observation(self):
        raise NotImplementedError()

    def render_update(self):
        #Paint God View
        self._screen.blit(self._surf_god, (self._view_size, 0))
        if(self.task_type == "SURVIVAL"):
            self.draw_food(self._screen, (self._view_size, 0))

        #Paint Agent and Observation
        self.render_observation()

        pygame.display.update()
        done = False
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                done=True
        keys = pygame.key.get_pressed()
        return done, keys

    def render_trajectory(self, file_name, additional=None):
        # Render god view with record on the trajectory
        if(additional is not None):
            aw, ah = additional["surfaces"][0].get_width(),additional["surfaces"][0].get_height()
        else:
            aw, ah = (0, 0)

        traj_screen = pygame.Surface((self._view_size + aw, max(self._view_size, ah)))
        traj_screen.fill(pygame.Color("white"))
        traj_screen.blit(self._surf_god, (0, 0))

        pygame.draw.rect(traj_screen, pygame.Color("red"), 
                (self._agent_grid[0] * self._render_cell_size, self._view_size - (self._agent_grid[1] + 1) * self._render_cell_size,
                self._render_cell_size, self._render_cell_size), width=0)
        if(self.task_type == "SURVIVAL"):
            self.draw_food(traj_screen, (0, 0))

        for i in range(len(self._agent_trajectory)-1):
            p = self._agent_trajectory[i]
            n = self._agent_trajectory[i+1]
            p = [(p[0] + 0.5) * self._render_cell_size, self._view_size - (p[1] + 0.5) *  self._render_cell_size]
            n = [(n[0] + 0.5) * self._render_cell_size, self._view_size - (n[1] + 0.5) *  self._render_cell_size]
            pygame.draw.line(traj_screen, pygame.Color("red"), p, n, width=3)

        # paint some additional surfaces where necessary
        if(additional != None):
            for i in range(len(additional["surfaces"])):
                traj_screen.blit(additional["surfaces"][i], (self._view_size, 0))
                pygame.image.save(traj_screen, file_name.split(".")[0] + additional["file_names"][i] + ".png")
        else:
            pygame.image.save(traj_screen, file_name)

    def episode_is_over(self):
        return self.steps > self.max_steps-1

    def get_cell_center(self, cell):
        p_x = cell[0] * self._cell_size + 0.5 * self._cell_size
        p_y = cell[1] * self._cell_size + 0.5 * self._cell_size
        return [p_x, p_y]

    def get_loc_grid(self, loc):
        p_x = int(loc[0] / self._cell_size)
        p_y = int(loc[1] / self._cell_size)
        return [p_x, p_y]

    def movement_control(self, keys):
        raise NotImplementedError()

    def update_observation(self):
        raise NotImplementedError()

    def get_observation(self):
        return numpy.copy(self._observation)



"""
Core File of Maze Env
"""
import os
import numpy
import pygame
import random
from collections import namedtuple
from numpy import random as npyrnd
from numpy.linalg import norm

class MazeTaskManager(object):

    # Configurations that decides a specific task
    TaskConfig = namedtuple("TaskConfig", ["start", "goal", "cell_walls", "cell_texts", 
        "cell_size", "wall_height", "agent_height", "initial_life", "max_life",
        "step_reward", "goal_reward", "food_rewards", "food_interval"])

    def __init__(self, texture_dir, verbose=False):
        pathes = os.path.split(os.path.abspath(__file__))
        texture_dir = os.sep.join([pathes[0], texture_dir])
        texture_files = os.listdir(texture_dir)
        texture_files.sort()
        grounds = [None]
        for file_name in texture_files:
            if(file_name.find("wall") >= 0):
                grounds.append(pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name]))))
            if(file_name.find("ground") >= 0):
                grounds[0] = pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name])))
            if(file_name.find("ceil") >= 0):
                self.ceil = pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name])))
            if(file_name.find("arrow") >= 0):
                self.arrow = pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name])))
        self.grounds = numpy.asarray(grounds, dtype="float32")
        self.verbose = verbose

    @property
    def n_texts(self):
        return self.grounds.shape[0]

    def sample_task(self,
            n=15, 
            allow_loops=True, 
            cell_size=2.0, 
            wall_height=3.2, 
            agent_height=1.6,
            step_reward=-0.01,
            goal_reward=None,
            food_reward=0.50,
            initial_life=1.0,
            max_life=2.0,
            food_density=0.010,
            food_interval=100,
            crowd_ratio=0.0):
        # Initialize the maze ...
        assert n > 6, "Minimum required cells are 7"
        assert n % 2 != 0, "Cell Numbers can only be odd"
        if(self.verbose):
            print("Generating an random maze of size %dx%d, with allow loops=%s, crowd ratio=%f"%(n, n, allow_loops, crowd_ratio))
        cell_walls = numpy.ones(shape=(n, n), dtype="int32")
        cell_texts = numpy.random.randint(1, self.n_texts, size=(n, n))

        # Dig the initial positions
        for i in range(1, n, 2):
            for j in range(1, n, 2):
                cell_walls[i,j] = 0
        #Randomize a start, around the bottom-left
        s_x = random.randint(0, (n - 1) // 2 - 1) * 2 + 1
        s_y = random.randint(0, (n - 1) // 2 - 1) * 2 + 1
        start = (s_x, s_y)

        #Randomize a goal (not quite efficiently)
        goal = (n - 2, n - 2)
        minimum_goal_dist = 0.45 * n #(at least that far)
        for e_x in range(1, n, 2):
            for e_y in range(1, n, 2):
                e_x = random.randint(0, (n - 1) // 2 - 1) * 2 + 1
                e_y = random.randint(0, (n - 1) // 2 - 1) * 2 + 1
                d_s_e = numpy.sqrt((e_x - s_x) ** 2 + (e_y - s_y) ** 2)
                if(d_s_e > minimum_goal_dist):
                    goal = (e_x, e_y)
                    break

        #Initialize the logics for prim based maze generation
        wall_dict = dict()
        path_dict = dict()
        rev_path_dict = dict()
        path_idx = 0
        for i in range(1, n - 1):
            for j in range(1, n - 1):
                if(cell_walls[i,j] > 0): # we will keep the axial point
                    wall_dict[i, j] = 0
                elif(cell_walls[i,j] == 0):
                    path_dict[i, j] = path_idx
                    rev_path_dict[path_idx] = [(i,j)]
                    path_idx += 1

        #Prim the wall until start and goal are connected
        #while path_dict[start] != path_dict[goal]:
        #Prim the wall until all points are connected
        max_cell_walls = numpy.product(cell_walls[1:-1, 1:-1].shape)
        while len(rev_path_dict) > 1 or (allow_loops and numpy.sum(cell_walls[1:-1, 1:-1]) > max_cell_walls * crowd_ratio):
            wall_list = list(wall_dict.keys())
            random.shuffle(wall_list)
            for i, j in wall_list:
                new_path_id = -1
                connected_path_id = dict()
                abandon_path_id = dict()
                max_duplicate = 1

                for d_i, d_j in [(i - 1, j), (i + 1, j), (i, j - 1), (i, j + 1)]:
                    if((d_i > 0  and d_i < n and d_j > 0 and d_j < n)
                            and cell_walls[d_i, d_j] < 1):
                        # calculate duplicate path id that might creat a loop
                        if path_dict[d_i, d_j] not in connected_path_id:
                            connected_path_id[path_dict[d_i, d_j]] = 1
                        else:
                            connected_path_id[path_dict[d_i, d_j]] += 1
                        if(connected_path_id[path_dict[d_i, d_j]] > max_duplicate):
                            max_duplicate = connected_path_id[path_dict[d_i, d_j]]

                        # decide the new path_id and find those to be deleted
                        if(path_dict[d_i, d_j] < new_path_id or new_path_id < 0):
                            if(new_path_id >= 0):
                                abandon_path_id[new_path_id] = (new_i, new_j)
                            new_path_id = path_dict[d_i, d_j]
                            new_i = d_i
                            new_j = d_j
                        elif(path_dict[d_i, d_j] != new_path_id): # need to be abandoned
                            abandon_path_id[path_dict[d_i, d_j]] = (d_i, d_j)
                if(len(abandon_path_id) >= 1 and max_duplicate < 2):
                    break
                if(len(abandon_path_id) >= 1 and max_duplicate > 1 and allow_loops):
                    break
                if(allow_loops and len(rev_path_dict) < 2 and random.random() < 0.2):
                    break

            if(new_path_id < 0):
                continue
                        
            # add the released wall
            rev_path_dict[new_path_id].append((i,j))
            path_dict[i,j] = new_path_id
            cell_walls[i,j] = 0
            del wall_dict[i,j]

            # merge the path
            for path_id in abandon_path_id:
                rev_path_dict[new_path_id].extend(rev_path_dict[path_id])
                for t_i_o, t_j_o in rev_path_dict[path_id]:
                    path_dict[t_i_o,t_j_o] = new_path_id
                del rev_path_dict[path_id]

        #Paint the texture of passways to ground textures 
        for i in range(1, n - 1):
            for j in range(1, n - 1):
                if(cell_walls[i,j] < 1):
                    cell_texts[i,j] = 0

        #Calculate goal reward, default is - n sqrt(n) * step_reward
        assert step_reward < 0, "step_reward must be < 0"
        if(goal_reward is None):
            def_goal_reward = - numpy.sqrt(n) * n * step_reward
        else:
            def_goal_reward = goal_reward
        assert def_goal_reward > 0, "goal reward must be > 0"

        # Generate food points
        food_rewards = numpy.clip(numpy.random.rand(n, n) * food_reward, 0.10, food_reward)
        food_rewards *= 1.0 - cell_walls
        exp_food = (n - 1) * (n - 1) * food_density
        while(numpy.sum(food_rewards) > exp_food): 
            food_rewards *= (numpy.random.rand(n, n) < 0.90).astype("float32")
        food_interval = food_interval * (food_rewards > 1.0e-3).astype("int32")

        return MazeTaskManager.TaskConfig(
                start=start,
                goal=goal,
                cell_walls=cell_walls,
                cell_texts=cell_texts,
                cell_size=cell_size,
                step_reward=step_reward,
                goal_reward=def_goal_reward,
                wall_height=wall_height,
                agent_height=agent_height,
                initial_life=initial_life,
                max_life=max_life,
                food_rewards=food_rewards,
                food_interval=food_interval
                )

MAZE_TASK_MANAGER=MazeTaskManager("img")
MazeTaskSampler = MAZE_TASK_MANAGER.sample_task
